use std::sync::OnceLock;

use image::{DynamicImage, GenericImageView, RgbaImage};
use thiserror::Error;
#[derive(Debug, Error)]
pub enum TextureError {
    #[error("Failed to load image")]
    ImageError(
        #[from]
        #[source]
        image::ImageError,
    ),
}
pub struct Texture {
    #[allow(unused)]
    pub texture: wgpu::Texture,
    pub view: wgpu::TextureView,
    pub sampler: wgpu::Sampler,
}
impl Texture {
    
    pub fn from_bytes(
        device: &wgpu::Device,
        queue: &wgpu::Queue,
        bytes: &[u8],
        label: &str,
    ) -> Result<Self, TextureError> {
        let img = image::load_from_memory(bytes)?;
        Self::from_image(device, queue, &img, Some(label))
    }

    pub fn from_image(
        device: &wgpu::Device,
        queue: &wgpu::Queue,
        img: &image::DynamicImage,
        label: Option<&str>,
    ) -> Result<Self, TextureError> {
        let rgba = img.to_rgba8();
        let dimensions = img.dimensions();

        let size = wgpu::Extent3d {
            width: dimensions.0,
            height: dimensions.1,
            depth_or_array_layers: 1,
        };
        let texture = device.create_texture(&wgpu::TextureDescriptor {
            label,
            size,
            mip_level_count: 1,
            sample_count: 1,
            dimension: wgpu::TextureDimension::D2,
            format: wgpu::TextureFormat::Rgba8UnormSrgb,
            usage: wgpu::TextureUsages::TEXTURE_BINDING | wgpu::TextureUsages::COPY_DST,
            view_formats: &[],
        });

        queue.write_texture(
            wgpu::TexelCopyTextureInfo {
                aspect: wgpu::TextureAspect::All,
                texture: &texture,
                mip_level: 0,
                origin: wgpu::Origin3d::ZERO,
            },
            &rgba,
            wgpu::TexelCopyBufferLayout {
                offset: 0,
                bytes_per_row: Some(4 * dimensions.0),
                rows_per_image: Some(dimensions.1),
            },
            size,
        );

        let view = texture.create_view(&wgpu::TextureViewDescriptor::default());
        let sampler = device.create_sampler(&wgpu::SamplerDescriptor {
            address_mode_u: wgpu::AddressMode::ClampToEdge,
            address_mode_v: wgpu::AddressMode::ClampToEdge,
            address_mode_w: wgpu::AddressMode::ClampToEdge,
            mag_filter: wgpu::FilterMode::Linear,
            min_filter: wgpu::FilterMode::Nearest,
            mipmap_filter: wgpu::FilterMode::Nearest,
            ..Default::default()
        });

        Ok(Self {
            texture,
            view,
            sampler,
        })
    }
}
static CHECKERBOARD_IMAGE: OnceLock<DynamicImage> = OnceLock::new();
pub fn create_checkerboard(size: u32, tile_size: u32) -> DynamicImage {
    let mut img = RgbaImage::new(size, size);
    let pixels = img.as_mut();
    let bytes_per_pixel = 4; // RGBA

    for y in 0..size {
        let tile_y = y / tile_size;
        for x in 0..size {
            let tile_x = x / tile_size;
            let color = if (tile_x + tile_y) % 2 == 0 {
                [0, 0, 0, 255] 
            } else {
                [255, 5, 255, 255]
            };
            let pixel_index = (y * size + x) as usize * bytes_per_pixel;
            pixels[pixel_index..pixel_index + bytes_per_pixel].copy_from_slice(&color);
        }
    }

    image::DynamicImage::ImageRgba8(img)
}
#[tracing::instrument]
pub fn get_checkerboard() -> &'static DynamicImage {
    CHECKERBOARD_IMAGE.get_or_init(|| {
        let size = 512;
        let tile_size = 24;
        create_checkerboard(size, tile_size)
    })
}
